@@ -30,8 +30,6 @@
 #include <gdk/gdkx.h>
 #endif
 
-#include <gconf/gconf.h>
-
 #include "terminal-accels.h"
 #include "terminal-app.h"
 #include "terminal-debug.h"
@@ -42,6 +40,11 @@
 #include "terminal-util.h"
 #include "terminal-window.h"
 
+#if GTK_CHECK_VERSION (2, 18, 0)
+#include "terminal-info-bar.h"
+#define USE_INFOBARS
+#endif
+
 #include "eggshell.h"
 
 #define URL_MATCH_CURSOR  (GDK_HAND2)
@@ -55,7 +58,6 @@ typedef struct
 
 struct _TerminalScreenPrivate
 {
-  TerminalWindow *window;
   TerminalProfile *profile; /* may be NULL at times */
   guint profile_changed_id;
   guint profile_forgotten_id;
@@ -150,7 +152,7 @@ static guint signals[LAST_SIGNAL];
 #define PORT "(?:\\:[[:digit:]]{1,5})?"
 #define PATHCHARS_CLASS "[-[:alnum:]\\Q_$.+!*,;@&=?/~#%\\E]"
 #define PATHTERM_CLASS "[^\\Q]'.}>) \t\r\n,\"\\E]"
-#define SCHEME "(?:news:|telnet:|nntp:|file:\\/|https?:|ftps?:|webcal:)"
+#define SCHEME "(?:news:|telnet:|nntp:|file:\\/|https?:|ftps?:|sftp:|webcal:)"
 #define USERPASS USERCHARS_CLASS "+(?:" PASSCHARS_CLASS "+)?"
 #define URLPATH   "(?:(/"PATHCHARS_CLASS"+(?:[(]"PATHCHARS_CLASS"*[)])*"PATHCHARS_CLASS"*)*"PATHTERM_CLASS")?"
 
@@ -240,62 +242,6 @@ free_tag_data (TagData *tagdata)
   g_slice_free (TagData, tagdata);
 }
 
-#ifdef GNOME_ENABLE_DEBUG
-static void
-parent_size_request (GtkWidget *scrolled_window, GtkRequisition *req, GtkWidget *screen)
-{
-  _terminal_debug_print (TERMINAL_DEBUG_GEOMETRY,
-                         "[screen %p] scrolled-window size req %d : %d\n",
-                         screen, req->width, req->height);
-}
-#endif
-
-static void
-parent_parent_set_cb (GtkWidget *widget,
-                      GtkWidget *old_parent,
-                      TerminalScreen *screen)
-{
-  TerminalScreenPrivate *priv = screen->priv;
-  GtkWidget *toplevel;
-
-  if (widget->parent)
-    {
-      g_return_if_fail (GTK_IS_NOTEBOOK (widget->parent));
-
-      toplevel = gtk_widget_get_toplevel (widget);
-#if GTK_CHECK_VERSION (2, 19, 3)
-      g_return_if_fail (gtk_widget_is_toplevel (toplevel));
-#else
-      g_return_if_fail (GTK_WIDGET_TOPLEVEL (toplevel));
-#endif
-
-      priv->window = TERMINAL_WINDOW (toplevel);
-    }
-  else
-    priv->window = NULL;
-}
-
-static void
-parent_set_callback (GtkWidget *widget,
-                     GtkWidget *old_parent)
-{
-  if (old_parent)
-    g_signal_handlers_disconnect_by_func (old_parent, G_CALLBACK (parent_parent_set_cb), widget);
-
-  if (widget->parent)
-    g_signal_connect (widget->parent, "parent-set", G_CALLBACK (parent_parent_set_cb), widget);
-
-#ifdef GNOME_ENABLE_DEBUG
-  _TERMINAL_DEBUG_IF (TERMINAL_DEBUG_GEOMETRY)
-    {
-      if (old_parent)
-        g_signal_handlers_disconnect_by_func (old_parent, G_CALLBACK (parent_size_request), widget);
-      if (widget->parent)
-        g_signal_connect (widget->parent, "size-request", G_CALLBACK (parent_size_request), widget);
-    }
-#endif
-}
-
 static void
 terminal_screen_class_enable_menu_bar_accel_notify_cb (TerminalApp *app,
                                                        GParamSpec *pspec,
@@ -320,6 +266,35 @@ terminal_screen_class_enable_menu_bar_accel_notify_cb (TerminalApp *app,
     gtk_binding_entry_skip (binding_set, GDK_F10, GDK_SHIFT_MASK);
 }
 
+static TerminalWindow *
+terminal_screen_get_window (TerminalScreen *screen)
+{
+  GtkWidget *widget = GTK_WIDGET (screen);
+  GtkWidget *toplevel;
+
+  toplevel = gtk_widget_get_toplevel (widget);
+#if GTK_CHECK_VERSION (2, 18, 0)
+  if (!gtk_widget_is_toplevel (toplevel))
+#else
+  if (!GTK_WIDGET_TOPLEVEL (toplevel))
+#endif
+    return NULL;
+
+  return TERMINAL_WINDOW (toplevel);
+}
+
+static gboolean
+window_uses_argb_visual (TerminalScreen *screen)
+{
+  TerminalWindow *window;
+
+  window = terminal_screen_get_window (screen);
+  if (window == NULL || !GTK_WIDGET_REALIZED (window))
+    return FALSE;
+
+  return terminal_window_uses_argb_visual (window);
+}
+
 static void
 terminal_screen_realize (GtkWidget *widget)
 {
@@ -329,13 +304,11 @@ terminal_screen_realize (GtkWidget *widget)
 
   GTK_WIDGET_CLASS (terminal_screen_parent_class)->realize (widget);
 
-  g_assert (priv->window != NULL);
-
   /* FIXME: Don't enable this if we have a compmgr. */
   bg_type = terminal_profile_get_property_enum (priv->profile, TERMINAL_PROFILE_BACKGROUND_TYPE);
   vte_terminal_set_background_transparent (VTE_TERMINAL (screen),
                                            bg_type == TERMINAL_BACKGROUND_TRANSPARENT &&
-                                           !terminal_window_uses_argb_visual (priv->window));
+                                           !window_uses_argb_visual (screen));
 }
 
 static void
@@ -444,8 +417,6 @@ terminal_screen_init (TerminalScreen *screen)
   g_signal_connect (terminal_app_get (), "notify::system-font",
                     G_CALLBACK (terminal_screen_system_font_notify_cb), screen);
 
-  g_signal_connect (screen, "parent-set", G_CALLBACK (parent_set_callback), NULL);
-
 #ifdef GNOME_ENABLE_DEBUG
   _TERMINAL_DEBUG_IF (TERMINAL_DEBUG_GEOMETRY)
     {
@@ -738,9 +709,11 @@ terminal_screen_new (TerminalProfile *profile,
 
   terminal_screen_set_profile (screen, profile);
 
-  vte_terminal_set_size (VTE_TERMINAL (screen),
-			 terminal_profile_get_property_int (profile, TERMINAL_PROFILE_DEFAULT_SIZE_COLUMNS),
-			 terminal_profile_get_property_int (profile, TERMINAL_PROFILE_DEFAULT_SIZE_ROWS));
+  if (terminal_profile_get_property_boolean (profile, TERMINAL_PROFILE_USE_CUSTOM_DEFAULT_SIZE)) {
+    vte_terminal_set_size (VTE_TERMINAL (screen),
+			   terminal_profile_get_property_int (profile, TERMINAL_PROFILE_DEFAULT_SIZE_COLUMNS),
+			   terminal_profile_get_property_int (profile, TERMINAL_PROFILE_DEFAULT_SIZE_ROWS));
+  }
 
   if (title)
     terminal_screen_set_override_title (screen, title);
@@ -937,6 +910,7 @@ terminal_screen_profile_notify_cb (TerminalProfile *profile,
   VteTerminal *vte_terminal = VTE_TERMINAL (screen);
   const char *prop_name;
   TerminalBackgroundType bg_type;
+  TerminalWindow *window;
 
   if (pspec)
     prop_name = pspec->name;
@@ -945,12 +919,12 @@ terminal_screen_profile_notify_cb (TerminalProfile *profile,
 
   g_object_freeze_notify (object);
 
-  if (priv->window)
+  if ((window = terminal_screen_get_window (screen)))
     {
       /* We need these in line for the set_size in
        * update_on_realize
        */
-      terminal_window_update_geometry (priv->window);
+      terminal_window_update_geometry (window);
     }
   
   if (!prop_name || prop_name == I_(TERMINAL_PROFILE_SCROLLBAR_POSITION))
@@ -1062,7 +1036,7 @@ terminal_screen_profile_notify_cb (TerminalProfile *profile,
       /* FIXME: Don't enable this if we have a compmgr. */
       vte_terminal_set_background_transparent (vte_terminal,
                                                bg_type == TERMINAL_BACKGROUND_TRANSPARENT &&
-                                               (!priv->window || !terminal_window_uses_argb_visual (priv->window)));
+                                               !window_uses_argb_visual (screen));
     }
 
   if (!prop_name || prop_name == I_(TERMINAL_PROFILE_BACKSPACE_BINDING))
@@ -1178,11 +1152,12 @@ terminal_screen_system_font_notify_cb (TerminalApp *app,
 static void
 terminal_screen_change_font (TerminalScreen *screen)
 {
-  TerminalScreenPrivate *priv = screen->priv;
+  TerminalWindow *window;
 
   terminal_screen_set_font (screen);
 
-  terminal_window_set_size (priv->window, screen, TRUE);
+  window = terminal_screen_get_window (screen);
+  terminal_window_set_size (window, screen, TRUE);
 }
 
 static void
@@ -1297,275 +1272,28 @@ terminal_screen_get_initial_environment (TerminalScreen *screen)
   return screen->priv->initial_env;
 }
 
-static void
-show_command_error_dialog (TerminalScreen *screen,
-                           GError         *error)
-{
-  g_assert (error != NULL);
-  
-  terminal_util_show_error_dialog ((GtkWindow*) gtk_widget_get_ancestor (GTK_WIDGET (screen), GTK_TYPE_WINDOW), NULL, error,
-                                   "%s", _("There was a problem with the command for this terminal"));
-}
-
-
-static char *
-conf_get_string (GConfClient *conf, const char *key)
-{
-  char *value;
-  value = gconf_client_get_string (conf, key, NULL);
-  if (G_UNLIKELY (value && *value == '\0'))
-    {
-      g_free (value);
-      value = NULL;
-    }
-  return value;
-}
-
-static gboolean
-conf_get_bool (GConfClient *conf, const char *key)
-{
-  return gconf_client_get_bool (conf, key, NULL);
-}
-
-static gint
-conf_get_int (GConfClient *conf, const char *key)
-{
-  return gconf_client_get_int (conf, key, NULL);
-}
-
-/* Consumes value.
- * Sets for both key and upper(key).
- * Also, never overwrites a variable. */
-static void
-set_proxy_env (GHashTable *env_table, const char *key, char *value)
-{
-  char *key1 = NULL, *key2 = NULL;
-  char *value1 = NULL, *value2 = NULL;
-
-  if (!value)
-    return;
-
-  if (g_hash_table_lookup (env_table, key) == NULL)
-    key1 = g_strdup (key);
-
-  key2 = g_ascii_strup (key, -1);
-  if (g_hash_table_lookup (env_table, key) != NULL)
-    {
-      g_free (key2);
-      key2 = NULL;
-    }
-
-  if (key1 && key2)
-    {
-      value1 = value;
-      value2 = g_strdup (value);
-    }
-  else if (key1)
-    value1 = value;
-  else if (key2)
-    value2 = value;
-  else
-    g_free (value);
-
-  if (key1)
-    g_hash_table_replace (env_table, key1, value1);
-  if (key2)
-    g_hash_table_replace (env_table, key2, value2);
-}
-
-
-static void
-setup_http_proxy_env (GHashTable *env_table, GConfClient *conf)
-{
-  gchar *host, *auth = NULL;
-  gint port;
-  GSList *ignore;
-
-#define HTTP_PROXY_DIR "/system/http_proxy"
-
-  gconf_client_preload (conf, HTTP_PROXY_DIR, GCONF_CLIENT_PRELOAD_ONELEVEL, NULL);
-
-  if (!conf_get_bool (conf, HTTP_PROXY_DIR "/use_http_proxy"))
-    return;
-
-  if (conf_get_bool (conf, HTTP_PROXY_DIR "/use_authentication"))
-    {
-      char *user, *password;
-      user = conf_get_string (conf, HTTP_PROXY_DIR "/authentication_user");
-      password = conf_get_string (conf, HTTP_PROXY_DIR "/authentication_password");
-      if (user)
-	{
-	  if (password)
-	    {
-	      auth = g_strdup_printf ("%s:%s", user, password);
-	      g_free (user);
-	    }
-	  else
-	    auth = user;
-	}
-      g_free (password);
-    }
-
-  host = conf_get_string (conf, HTTP_PROXY_DIR "/host");
-  port = conf_get_int (conf, HTTP_PROXY_DIR "/port");
-  if (host && port)
-    {
-      char *proxy;
-      if (auth)
-	proxy = g_strdup_printf ("http://%s@%s:%d/", auth, host, port);
-      else
-	proxy = g_strdup_printf ("http://%s:%d/", host, port);
-      set_proxy_env (env_table, "http_proxy", proxy);
-    }
-  g_free (host);
-
-  g_free (auth);
-
-
-  ignore = gconf_client_get_list (conf, HTTP_PROXY_DIR "/ignore_hosts", GCONF_VALUE_STRING, NULL);
-  if (ignore)
-    {
-      GString *buf = g_string_sized_new (64);
-      while (ignore != NULL)
-	{
-	  GSList *old;
-
-	  g_string_append (buf, ignore->data);
-	  g_string_append_c (buf, ',');
-
-	  old = ignore;
-	  ignore = g_slist_next (ignore);
-	  g_free (old->data);
-	  g_slist_free_1 (old);
-	}
-      set_proxy_env (env_table, "no_proxy", g_string_free (buf, FALSE));
-    }
-}
-
-#define PROXY_DIR "/system/proxy"
-
-static void
-setup_https_proxy_env (GHashTable *env_table, GConfClient *conf)
-{
-  gchar *host;
-  gint port;
-
-  host = conf_get_string (conf, PROXY_DIR "/secure_host");
-  port = conf_get_int (conf, PROXY_DIR "/secure_port");
-  if (host && port)
-    {
-      char *proxy;
-      proxy = g_strdup_printf ("https://%s:%d/", host, port);
-      set_proxy_env (env_table, "https_proxy", proxy);
-    }
-  g_free (host);
-}
-
-static void
-setup_ftp_proxy_env (GHashTable *env_table, GConfClient *conf)
-{
-  gchar *host;
-  gint port;
-
-  host = conf_get_string (conf, PROXY_DIR "/ftp_host");
-  port = conf_get_int (conf, PROXY_DIR "/ftp_port");
-  if (host && port)
-    {
-      char *proxy;
-      proxy = g_strdup_printf ("ftp://%s:%d/", host, port);
-      set_proxy_env (env_table, "ftp_proxy", proxy);
-    }
-  g_free (host);
-}
-
-static void
-setup_socks_proxy_env (GHashTable *env_table, GConfClient *conf)
-{
-  gchar *host;
-  gint port;
-
-  host = conf_get_string (conf, PROXY_DIR "/socks_host");
-  port = conf_get_int (conf, PROXY_DIR "/socks_port");
-  if (host && port)
-    {
-      char *proxy;
-      proxy = g_strdup_printf ("socks://%s:%d/", host, port);
-      set_proxy_env (env_table, "all_proxy", proxy);
-    }
-  g_free (host);
-}
-
-static void
-setup_autoconfig_proxy_env (GHashTable *env_table, GConfClient *conf)
-{
-  gchar *url;
-
-  url = conf_get_string (conf, PROXY_DIR "/autoconfig_url");
-  if (url)
-    {
-      /* XXX  Not sure what to do with it.  See bug 596688
-      char *proxy;
-      proxy = g_strdup_printf ("pac+%s", url);
-      set_proxy_env (env_table, "http_proxy", proxy);
-      */
-    }
-  g_free (url);
-}
-
-static void
-setup_proxy_env (GHashTable *env_table)
-{
-  char *proxymode;
-
-  GConfClient *conf;
-  conf = gconf_client_get_default ();
-  gconf_client_preload (conf, PROXY_DIR, GCONF_CLIENT_PRELOAD_ONELEVEL, NULL);
-
-  /* If mode is not manual, nothing to set */
-  proxymode = conf_get_string (conf, PROXY_DIR "/mode");
-  if (proxymode && 0 == strcmp (proxymode, "manual"))
-    {
-      setup_http_proxy_env (env_table, conf);
-      setup_https_proxy_env (env_table, conf);
-      setup_ftp_proxy_env (env_table, conf);
-      setup_socks_proxy_env (env_table, conf);
-    }
-  else if (proxymode && 0 == strcmp (proxymode, "auto"))
-    {
-      setup_autoconfig_proxy_env (env_table, conf);
-    }
-
-  g_free (proxymode);
-  g_object_unref (conf);
-}
-
-
 static gboolean
 get_child_command (TerminalScreen *screen,
                    const char     *shell_env,
-                   char          **file_p,
+                   GSpawnFlags    *spawn_flags_p,
                    char         ***argv_p,
                    GError        **err)
 {
   TerminalScreenPrivate *priv = screen->priv;
   TerminalProfile *profile;
-  char  *file;
   char **argv;
 
+  g_assert (spawn_flags_p != NULL && argv_p != NULL);
+
   profile = priv->profile;
 
-  file = NULL;
-  argv = NULL;
-  
-  if (file_p)
-    *file_p = NULL;
-  if (argv_p)
-    *argv_p = NULL;
+  *argv_p = argv = NULL;
 
   if (priv->override_command)
     {
-      file = g_strdup (priv->override_command[0]);
       argv = g_strdupv (priv->override_command);
+
+      *spawn_flags_p |= G_SPAWN_SEARCH_PATH;
     }
   else if (terminal_profile_get_property_boolean (profile, TERMINAL_PROFILE_USE_CUSTOM_COMMAND))
     {
@@ -1574,44 +1302,37 @@ get_child_command (TerminalScreen *screen,
                                err))
         return FALSE;
 
-      file = g_strdup (argv[0]);
+      *spawn_flags_p |= G_SPAWN_SEARCH_PATH;
     }
   else
     {
       const char *only_name;
       char *shell;
+      int argc = 0;
 
       shell = egg_shell (shell_env);
 
-      file = g_strdup (shell);
-      
       only_name = strrchr (shell, '/');
       if (only_name != NULL)
         only_name++;
       else
         only_name = shell;
 
-      argv = g_new (char*, 2);
+      argv = g_new (char*, 3);
+
+      argv[argc++] = shell;
 
       if (terminal_profile_get_property_boolean (profile, TERMINAL_PROFILE_LOGIN_SHELL))
-        argv[0] = g_strconcat ("-", only_name, NULL);
+        argv[argc++] = g_strconcat ("-", only_name, NULL);
       else
-        argv[0] = g_strdup (only_name);
+        argv[argc++] = g_strdup (only_name);
 
-      argv[1] = NULL;
+      argv[argc++] = NULL;
 
-      g_free (shell);
+      *spawn_flags_p |= G_SPAWN_FILE_AND_ARGV_ZERO;
     }
 
-  if (file_p)
-    *file_p = file;
-  else
-    g_free (file);
-
-  if (argv_p)
-    *argv_p = argv;
-  else
-    g_strfreev (argv);
+  *argv_p = argv;
 
   return TRUE;
 }
@@ -1673,7 +1394,7 @@ get_child_environment (TerminalScreen *screen,
   g_hash_table_replace (env_table, g_strdup ("DISPLAY"), g_strdup (gdk_display_get_name (gtk_widget_get_display (window))));
 #endif
 
-  setup_proxy_env (env_table);
+  terminal_util_add_proxy_env (env_table);
 
   retval = g_ptr_array_sized_new (g_hash_table_size (env_table));
   g_hash_table_iter_init (&iter, env_table);
@@ -1687,6 +1408,43 @@ get_child_environment (TerminalScreen *screen,
   return (char **) g_ptr_array_free (retval, FALSE);
 }
 
+#ifdef USE_INFOBARS
+
+enum {
+  RESPONSE_RELAUNCH,
+  RESPONSE_EDIT_PROFILE
+};
+
+static void
+info_bar_response_cb (GtkWidget *info_bar,
+                      int response,
+                      TerminalScreen *screen)
+{
+  gtk_widget_grab_focus (GTK_WIDGET (screen));
+
+  switch (response) {
+    case GTK_RESPONSE_CANCEL:
+      gtk_widget_destroy (info_bar);
+      g_signal_emit (screen, signals[CLOSE_SCREEN], 0);
+      break;
+    case RESPONSE_RELAUNCH:
+      gtk_widget_destroy (info_bar);
+      terminal_screen_launch_child_on_idle (screen);
+      break;
+    case RESPONSE_EDIT_PROFILE:
+      terminal_app_edit_profile (terminal_app_get (),
+                                 terminal_screen_get_profile (screen),
+                                 GTK_WINDOW (terminal_screen_get_window (screen)),
+                                 "custom-command-entry");
+      break;
+    default:
+      gtk_widget_destroy (info_bar);
+      break;
+  }
+}
+
+#endif /* USE_INFOBARS */
+
 static gboolean
 terminal_screen_launch_child_cb (TerminalScreen *screen)
 {
@@ -1694,10 +1452,12 @@ terminal_screen_launch_child_cb (TerminalScreen *screen)
   VteTerminal *terminal = VTE_TERMINAL (screen);
   TerminalProfile *profile;
   char **env, **argv;
-  char *path, *shell = NULL;
+  char *shell = NULL;
   GError *err = NULL;
-  gboolean update_records;
   const char *working_dir;
+  VtePtyFlags pty_flags = VTE_PTY_DEFAULT;
+  GSpawnFlags spawn_flags = 0;
+  GPid pid;
 
   priv->launch_child_source_id = 0;
 
@@ -1709,45 +1469,61 @@ terminal_screen_launch_child_cb (TerminalScreen *screen)
 
   env = get_child_environment (screen, &shell);
 
-  if (!get_child_command (screen, shell, &path, &argv, &err))
-    {
-      show_command_error_dialog (screen, err);
-      g_error_free (err);
-
-      g_strfreev (env);
-      g_free (shell);
-
-      return FALSE;
-    }
-
-  update_records = terminal_profile_get_property_boolean (profile, TERMINAL_PROFILE_UPDATE_RECORDS);
-
   if (priv->initial_working_directory)
     working_dir = priv->initial_working_directory;
   else
     working_dir = g_get_home_dir ();
 
-  priv->child_pid = vte_terminal_fork_command (terminal,
-                                               path,
-                                               argv,
-                                               env,
-                                               working_dir,
-                                               terminal_profile_get_property_boolean (profile, TERMINAL_PROFILE_LOGIN_SHELL),
-                                               update_records,
-                                               update_records);
+  if (!terminal_profile_get_property_boolean (profile, TERMINAL_PROFILE_LOGIN_SHELL))
+    pty_flags |= VTE_PTY_NO_LASTLOG;
+  if (!terminal_profile_get_property_boolean (profile, TERMINAL_PROFILE_UPDATE_RECORDS))
+    pty_flags |= VTE_PTY_NO_UTMP | VTE_PTY_NO_WTMP;
 
-  if (priv->child_pid == -1)
-    {
+  if (!get_child_command (screen, shell, &spawn_flags, &argv, &err) ||
+      !vte_terminal_fork_command_full (terminal,
+                                       pty_flags,
+                                       working_dir,
+                                       argv,
+                                       env,
+                                       spawn_flags,
+                                       NULL, NULL,
+                                       &pid,
+                                       &err)) {
+#ifdef USE_INFOBARS
+    GtkWidget *info_bar;
 
-      terminal_util_show_error_dialog ((GtkWindow*) gtk_widget_get_ancestor (GTK_WIDGET (screen), GTK_TYPE_WINDOW), NULL,
-                                       err,
-                                       "%s", _("There was an error creating the child process for this terminal"));
-    }
-  
+    info_bar = terminal_info_bar_new (GTK_MESSAGE_ERROR,
+                                      _("_Profile Preferences"), RESPONSE_EDIT_PROFILE,
+                                      _("_Relaunch"), RESPONSE_RELAUNCH,
+                                      NULL);
+    terminal_info_bar_format_text (TERMINAL_INFO_BAR (info_bar),
+                                   _("There was an error creating the child process for this terminal"));
+    terminal_info_bar_format_text (TERMINAL_INFO_BAR (info_bar),
+                                   "%s", err->message);
+    g_signal_connect (info_bar, "response",
+                      G_CALLBACK (info_bar_response_cb), screen);
+
+    gtk_box_pack_start (GTK_BOX (terminal_screen_container_get_from_screen (screen)),
+                        info_bar, FALSE, FALSE, 0);
+    gtk_info_bar_set_default_response (GTK_INFO_BAR (info_bar), GTK_RESPONSE_CANCEL);
+    gtk_widget_show (info_bar);
+#else
+    terminal_util_show_error_dialog (GTK_WINDOW (terminal_screen_get_window (screen)), NULL,
+                                     err,
+                                     "%s", _("There was an error creating the child process for this terminal"));
+#endif /* USE_INFOBARS */
+
+    g_error_free (err);
+    g_strfreev (env);
+    g_free (shell);
+
+    return FALSE;
+  }
+
+  priv->child_pid = pid;
   priv->pty_fd = vte_terminal_get_pty (terminal);
 
   g_free (shell);
-  g_free (path);
   g_strfreev (argv);
   g_strfreev (env);
 
@@ -1772,13 +1548,12 @@ terminal_screen_launch_child_on_idle (TerminalScreen *screen)
 static TerminalScreenPopupInfo *
 terminal_screen_popup_info_new (TerminalScreen *screen)
 {
-  TerminalScreenPrivate *priv = screen->priv;
   TerminalScreenPopupInfo *info;
 
   info = g_slice_new0 (TerminalScreenPopupInfo);
   info->ref_count = 1;
   info->screen = g_object_ref (screen);
-  info->window = priv->window;
+  info->window = terminal_screen_get_window (screen);
 
   return info;
 }
@@ -2103,7 +1878,37 @@ terminal_screen_child_exited (VteTerminal *terminal)
     case TERMINAL_EXIT_RESTART:
       terminal_screen_launch_child_on_idle (screen);
       break;
-    case TERMINAL_EXIT_HOLD:
+#ifdef USE_INFOBARS
+    case TERMINAL_EXIT_HOLD: {
+      GtkWidget *info_bar;
+      int status;
+
+      status = vte_terminal_get_child_exit_status (terminal);
+
+      info_bar = terminal_info_bar_new (GTK_MESSAGE_INFO,
+                                        _("_Relaunch"), RESPONSE_RELAUNCH,
+                                        NULL);
+      if (WIFEXITED (status)) {
+        terminal_info_bar_format_text (TERMINAL_INFO_BAR (info_bar),
+                                      _("The child process exited normally with status %d."), WEXITSTATUS (status));
+      } else if (WIFSIGNALED (status)) {
+        terminal_info_bar_format_text (TERMINAL_INFO_BAR (info_bar),
+                                      _("The child process was terminated by signal %d."), WTERMSIG (status));
+      } else {
+        terminal_info_bar_format_text (TERMINAL_INFO_BAR (info_bar),
+                                      _("The child process was terminated."));
+      }
+      g_signal_connect (info_bar, "response",
+                        G_CALLBACK (info_bar_response_cb), screen);
+
+      gtk_box_pack_start (GTK_BOX (terminal_screen_container_get_from_screen (screen)),
+                          info_bar, FALSE, FALSE, 0);
+      gtk_info_bar_set_default_response (GTK_INFO_BAR (info_bar), RESPONSE_RELAUNCH);
+      gtk_widget_show (info_bar);
+      break;
+    }
+#endif /* USE_INFOBARS */
+
     default:
       break;
     }
@@ -2329,13 +2134,13 @@ terminal_screen_drag_data_received (GtkWidget        *widget,
         if (!GTK_IS_WIDGET (container))
           return;
 
-        moving_screen = terminal_screen_container_get_screen (container);
+        moving_screen = terminal_screen_container_get_screen (TERMINAL_SCREEN_CONTAINER (container));
         g_return_if_fail (TERMINAL_IS_SCREEN (moving_screen));
         if (!TERMINAL_IS_SCREEN (moving_screen))
           return;
 
-        source_window = moving_screen->priv->window;
-        dest_window = screen->priv->window;
+        source_window = terminal_screen_get_window (moving_screen);
+        dest_window = terminal_screen_get_window (screen);
         dest_notebook = terminal_window_get_notebook (dest_window);
         page_num = gtk_notebook_page_num (GTK_NOTEBOOK (dest_notebook), 
                                           GTK_WIDGET (screen));
@@ -2354,12 +2159,12 @@ void
 _terminal_screen_update_scrollbar (TerminalScreen *screen)
 {
   TerminalScreenPrivate *priv = screen->priv;
-  GtkWidget *parent;
+  TerminalScreenContainer *container;
   GtkPolicyType policy = GTK_POLICY_ALWAYS;
   GtkCornerType corner = GTK_CORNER_TOP_LEFT;
 
-  parent = GTK_WIDGET (screen)->parent;
-  if (!parent)
+  container = terminal_screen_container_get_from_screen (screen);
+  if (container == NULL)
     return;
 
   switch (terminal_profile_get_property_enum (priv->profile, TERMINAL_PROFILE_SCROLLBAR_POSITION))
@@ -2380,8 +2185,8 @@ _terminal_screen_update_scrollbar (TerminalScreen *screen)
       break;
     }
 
-  terminal_screen_container_set_placement (parent, corner);
-  terminal_screen_container_set_policy (parent, GTK_POLICY_NEVER, policy);
+  terminal_screen_container_set_placement (container, corner);
+  terminal_screen_container_set_policy (container, GTK_POLICY_NEVER, policy);
 }
 
 void
