@@ -60,8 +60,6 @@ enum
 	PROP_WINDOW
 };
 
-static void	terminal_tabs_menu_class_init	(TerminalTabsMenuClass *klass);
-static void	terminal_tabs_menu_init	  	(TerminalTabsMenu *menu);
 static void	terminal_tabs_menu_update		(TerminalTabsMenu *menu);
 
 /* FIXME: this can be severely optimised */
@@ -130,7 +128,7 @@ free_tab_id (GtkAction *action)
         name = gtk_action_get_name (action);
         id = g_ascii_strtoull (name + ACTION_VERB_FORMAT_PREFIX_LEN, NULL,
                                ACTION_VERB_FORMAT_BASE);
-        g_assert (id >= 0 && id < tabs_id_array->len * 8);
+        g_assert (id < tabs_id_array->len * 8);
 
         b = id >> 3;
         bit = id & 0x7;
@@ -181,7 +179,7 @@ sync_tab_title (TerminalScreen *screen,
 
 static void
 notebook_page_added_cb (GtkNotebook *notebook,
-                        GtkWidget *container,
+                        TerminalScreenContainer *container,
 			guint position,
 			TerminalTabsMenu *menu)
 {
@@ -229,7 +227,7 @@ notebook_page_added_cb (GtkNotebook *notebook,
 
 static void
 notebook_page_removed_cb (GtkNotebook *notebook,
-                          GtkWidget *container,
+                          TerminalScreenContainer *container,
 			  guint position,
 			  TerminalTabsMenu *menu)
 {
@@ -271,11 +269,11 @@ notebook_page_switch_cb (GtkNotebook *notebook,
                          guint position,
                          TerminalTabsMenu *menu)
 {
-        GtkWidget *container;
+        TerminalScreenContainer *container;
         TerminalScreen *screen;
         GtkAction *action;
 
-        container = gtk_notebook_get_nth_page (notebook, position);
+        container = TERMINAL_SCREEN_CONTAINER (gtk_notebook_get_nth_page (notebook, position));
         screen = terminal_screen_container_get_screen (container);
 
 	action = g_object_get_data (G_OBJECT (screen), DATA_KEY);
@@ -350,6 +348,9 @@ terminal_tabs_menu_set_property (GObject *object,
 		case PROP_WINDOW:
 			terminal_tabs_menu_set_window (menu, g_value_get_object (value));
 			break;
+                default:
+                        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+                        break;
 	}
 }
 
@@ -458,7 +459,7 @@ terminal_tabs_menu_update (TerminalTabsMenu *menu)
 
 	for (l = tabs; l != NULL; l = l->next)
 	{
-                GtkWidget *container = l->data;
+                TerminalScreenContainer *container = TERMINAL_SCREEN_CONTAINER (l->data);
                 GObject *screen = G_OBJECT (terminal_screen_container_get_screen (container));
 
 		action = g_object_get_data (screen, DATA_KEY);
